import sys as _sys
import traceback as _traceback

import neaty.log as LOG

from pykbrd import app as app
from pykbrd import out as out
from pykbrd import kbrd as kbrd


class ExternalError(RuntimeError):
    pass


class InternalError(RuntimeError):
    pass


class UsageError(RuntimeError):
    pass


def _handle_exc(exc=None, es=5):
    """
    Handle exception *exc* and exit with status *es*.

    If *exc* is not provided, handle the Python's global exception--this
    should be whatever exception was thrown the last.

    If *es* is not provided, exit with status 5.

    This is to make sure exceptions are handled according to command line
    conventions.  Known exceptions such as usage errors, config errors or
    server errors should lead to warning and an intentional exit status.
    Any un-known exceptions will still show full traceback, but respect
    the app's logging mechanism and use specific exit status.
    """
    if exc:
        LOG.warn(str(exc))
    else:
        LOG.warn(''.join(_traceback.format_exc()))
    _sys.exit(es)


def _run(action, app):
    try:
        return app.run_action(action)
    except ExternalError as e:
        _handle_exc(exc=e, es=3)
    except InternalError as e:
        _handle_exc(exc=e, es=4)
    except Exception:
        _handle_exc()


def _init_app():
    """
    Initialize main application instance (app.App) from sys.argv
    """
    try:
        return app.App.from_argv(_sys.argv)
    except UsageError as e:
        _handle_exc(exc=e, es=2)
    except Exception:
        _handle_exc()


def _init_action(cls, scheme,
                 args,
                 ki=None,
                 base_params=None,
                 formatter_cls=None):
    """
    Initialize action instance for sub-command based on action arguments.

    *cls* must be descendant of app._BaseAction.

    Rest of the arguments are passed to the new() method of the class
    specified by *cls*.
    """
    try:
        return cls.new(
            ki=ki,
            scheme=scheme,
            base_params=base_params,
            args=args,
            formatter_cls=formatter_cls,
        )
    except UsageError as e:
        _handle_exc(exc=e, es=2)
    except Exception:
        _handle_exc()


def run(action_cls, scheme,
        ki=None,
        base_params=None,
        formatter_cls=None):
    """
    Instantiate and run action, handling CLI args and exceptions
    """
    app_, action_args = _init_app()
    action = _init_action(
        cls=action_cls,
        ki=ki,
        scheme=scheme,
        args=action_args,
        base_params=base_params,
        formatter_cls=formatter_cls,
    )
    return _run(action, app_)


__all__ = [
    ExternalError,
    InternalError,
    UsageError,
    app,
    out,
    run,
]
