class TableError(IndexError):
    pass


class Table:

    @classmethod
    def from_data(self, rows, wpadding=1, hpadding=0):
        trows = []
        for r in rows:
            trows.append(TableRow([
                TableCell(c) for c
                in r
            ]))
        return Table(trows, wpadding, hpadding)

    def __init__(self, rows, wpadding=1, hpadding=0):
        self.rows = rows
        for r in self.rows:
            r._table = self
        self.wpadding = wpadding
        self.hpadding = hpadding

    @property
    def width(self):
        return max([r.width for r in self.rows])

    @property
    def height(self):
        return len(self.rows)

    def column(self, n):
        cells = []
        if n >= self.width:
            raise TableError('no such column: %s' % n)
        for row in self.rows:
            try:
                cell = row.cells[n]
            except IndexError:
                cell = TableCell()
            cells.append(cell)
        return TableColumn(cells)

    def columns(self):
        columns = []
        for n in range(0, self.width - 1):
            columns.append(self.column(n))
        return columns

    def row(self, n):
        try:
            row = self.rows[n]
        except IndexError:
            raise TableError("no such row: %s" % n)
        return row

    def fmt(self):
        lines = []
        for row in self.rows:
            lines.extend(row.fmt())
        return lines


class TableCell:

    def __init__(self, data=None):
        self.data = data
        self._table = None

    def __str__(self):
        return str(self.data)

    @property
    def dheight(self):
        return len(str(self).splitlines())

    @property
    def dwidth(self):
        return len(str(self))

    @property
    def height(self):
        return 1

    @property
    def width(self):
        return 1

    def fmt(self, w, h):
        if w < self.dwidth:
            raise TableError('cannot fit into width: %d' % w)
        if h < self.dheight:
            raise TableError('cannot fit into height: %d' % h)
        lines = []
        togo = str(self).splitlines()
        while h:
            if togo:
                lines.append(('%%-%ds' % w) % togo.pop(0))
            else:
                lines.append('')
            h = h - 1
        return lines


class TableRow:

    def __init__(self, cells):
        self.cells = cells
        self._table = None
        for c in self.cells:
            c._table = self

    @property
    def cheight(self):
        return max([c.dheight for c in self.cells])

    @property
    def cwidth(self):
        return sum([c.dwidth for c in self.cells]) + self.width - 1

    @property
    def height(self):
        return 1

    @property
    def width(self):
        return len(self.cells)

    def fmt(self):
        ch = self.cheight + self._table.hpadding
        lines = [''] * ch
        for ci in range(self.width):
            cw = self._table.column(ci).cwidth + self._table.wpadding
            lno = 0
            for cline in self.cells[ci].fmt(cw, ch):
                lines[lno] = lines[lno] + cline
                lno = lno + 1
        return lines


class TableColumn:

    def __init__(self, cells):
        self.cells = cells

    @property
    def cheight(self):
        return sum([c.dheight for c in self.cells])

    @property
    def cwidth(self):
        return max([c.dwidth for c in self.cells])

    @property
    def height(self):
        return len(self.cells)

    @property
    def width(self):
        return 1
